from __future__ import annotations

from typing import Union

import yul.yul_ast as ast
from yul.AstMapper import AstMapper


class FoldIf(AstMapper):
    """This class replaces an if statment with its body or else_body if the
    condition is a literal.
    """

    def visit_block(self, node: ast.Block) -> ast.Block:
        statements = []
        for statement in node.statements:
            if isinstance(statement, ast.If):
                visited_if = self.visit(statement)
                if isinstance(visited_if, ast.Block):
                    statements += list(visited_if.statements)
                else:
                    statements.append(visited_if)
            else:
                statements.append(self.visit(statement))

        return ast.Block(tuple(statements))

    def visit_if(self, node: ast.If) -> Union[ast.Block, ast.If]:
        if isinstance(node.condition, ast.Literal):
            if node.condition.value:
                return self.visit(node.body)
            else:
                return (
                    self.visit(node.else_body) if node.else_body else ast.Block(tuple())
                )

        return super().visit_if(node)



import json
import os
import subprocess
from typing import Any, Dict, Optional, Sequence, Union

import aiohttp
from cli.encoding import get_cairo_calldata, get_ctor_evm_calldata, get_evm_calldata
from starkware.starknet.services.api.gateway.transaction import Transaction

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../.."))


async def send_req(method, url, tx: Optional[Union[str, Dict[str, Any]]] = None):
    if tx is not None:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method, url=url, data=Transaction.Schema().dumps(obj=tx)
            ) as response:
                text = await response.text()
                return text
    else:
        async with aiohttp.ClientSession() as session:
            async with session.request(method=method, url=url, data=None) as response:
                text = await response.text()
                return text


# returns true/false on transaction success/failure
async def _invoke(
    contract_base, program_info: dict, address, function, evm_inputs, network: str
):
    evm_calldata = get_evm_calldata(program_info["sol_abi"], function, evm_inputs)
    cairo_calldata = get_cairo_calldata(evm_calldata)
    starknet_invoke(contract_base, address, cairo_calldata, network)
    return True


def starknet_invoke(
    contract_base, address, cairo_calldata: Sequence[int], network: str
):
    abi = f"{contract_base}_abi.json"
    inputs = " ".join(map(str, cairo_calldata))
    print(
        os.popen(
            f"starknet invoke "
            f"--address {address} "
            f"--abi {abi} "
            f"--function __main "
            f"--inputs {inputs} "
            f"--network {network} "
        ).read()
    )


def starknet_compile(cairo_path, contract_base):
    compiled = f"{contract_base}_compiled.json"
    abi = f"{contract_base}_abi.json"
    process = subprocess.Popen(
        [
            "starknet-compile",
            cairo_path,
            "--output",
            compiled,
            "--abi",
            abi,
            "--cairo_path",
            f"{WARP_ROOT}/cairo-src",
        ]
    )
    output = process.wait()
    if output == 1:
        raise Exception("Compilation failed")
    return compiled


async def _deploy(
    cairo_path, contract_base, program_info, constructor_args, network: str
):
    evm_calldata = get_ctor_evm_calldata(program_info["sol_abi"], constructor_args)
    cairo_calldata = get_cairo_calldata(evm_calldata)
    starknet_deploy(contract_base, cairo_path, cairo_calldata, network)


def starknet_deploy(
    contract_base, cairo_path, cairo_calldata: Sequence[int], network: str
):
    compiled_contract = starknet_compile(cairo_path, contract_base)
    inputs = " ".join(map(str, cairo_calldata))
    print(
        os.popen(
            f"starknet deploy "
            f"--contract {contract_base}_compiled.json "
            f"--inputs {inputs} "
            f"--network {network} "
        ).read()
    )

    return compiled_contract


async def _status(tx_hash):
    status = f"https://alpha3.starknet.io/feeder_gateway/get_transaction_status?transactionHash={tx_hash}"
    res = await send_req("GET", status)
    print(json.loads(res))



from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from typing import Callable, Optional

import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor
from yul.BuiltinHandler import BuiltinHandler
from yul.FunctionGenerator import CairoFunctions
from yul.implicits import (
    IMPLICITS_SET,
    MANUAL_IMPLICITS,
    finalize_manual_implicit,
    initialize_manual_implicit,
    print_implicit,
)
from yul.Imports import format_imports, merge_imports
from yul.NameGenerator import NameGenerator
from yul.storage_access import generate_storage_var_declaration

UINT128_BOUND = 2 ** 128

COMMON_IMPORTS = {
    "starkware.cairo.common.registers": {"get_fp_and_pc"},
    "starkware.cairo.common.dict_access": {"DictAccess"},
    "starkware.cairo.common.default_dict": {
        "default_dict_new",
        "default_dict_finalize",
    },
    "starkware.cairo.common.uint256": {"Uint256"},
    "starkware.cairo.common.cairo_builtins": {
        "HashBuiltin",
        "BitwiseBuiltin",
    },
    "evm.array": {"validate_array"},
    "evm.exec_env": {"ExecutionEnvironment"},
}

MAIN_PREAMBLE = """%lang starknet
%builtins pedersen range_check bitwise
"""


class ToCairoVisitor(AstVisitor):
    def __init__(
        self,
        name_gen: NameGenerator,
        cairo_functions: CairoFunctions,
        builtins_map: Callable[[CairoFunctions], dict[str, BuiltinHandler]],
    ):
        super().__init__()
        self.name_gen = name_gen
        self.cairo_functions = cairo_functions
        self.builtins_map = builtins_map(self.cairo_functions)
        self.imports: defaultdict[str, set[str]] = defaultdict(set)
        merge_imports(self.imports, COMMON_IMPORTS)
        self.last_function: Optional[ast.FunctionDefinition] = None
        self.last_used_implicits: tuple[str, ...] = ()
        self.function_to_implicits: dict[str, set[str]] = {}
        self.next_stmt_is_leave: bool = False

    def translate(self, node: ast.Node) -> str:
        main_part = self.print(node)
        storage_vars = self.cairo_functions.storage_vars
        storage_var_decls = [
            generate_storage_var_declaration(x) for x in sorted(storage_vars)
        ]
        return "\n".join(
            [
                MAIN_PREAMBLE,
                format_imports(self.imports),
                "",
                *self.cairo_functions.get_definitions(),
                "",
                *storage_var_decls,
                self._make_constructor(),
                self._make_main(),
                main_part,
            ]
        )

    def print(self, node: ast.Node, *args, **kwargs) -> str:
        return self.visit(node, *args, **kwargs)

    def common_visit(self, node: ast.Node, *args, **kwargs):
        raise AssertionError(
            f"Each node type should have a custom visit, but {type(node)} doesn't"
        )

    def visit_typed_name(self, node: ast.TypedName) -> str:
        return f"{node.name} : {node.type}"

    def visit_literal(self, node: ast.Literal) -> str:
        v = int(node.value)  # to convert bools: True -> 1, False -> 0
        high, low = divmod(v, UINT128_BOUND)
        return f"Uint256(low={low}, high={high})"

    def visit_identifier(self, node: ast.Identifier) -> str:
        return f"{node.name}"

    def visit_assignment(self, node: ast.Assignment) -> str:
        return self.visit(
            ast.VariableDeclaration(
                variables=[ast.TypedName(x.name) for x in node.variable_names],
                value=node.value,
            )
        )

    def visit_function_call(self, node: ast.FunctionCall) -> str:
        fun_repr = self.print(node.function_name)
        args = [self.print(x) for x in node.arguments]
        if fun_repr == "revert":
            return "assert 0 = 1\njmp rel 0"
        if fun_repr == "pop":
            return ""
        result: str
        builtin_handler = self.builtins_map.get(fun_repr)
        if builtin_handler:
            result = builtin_handler.get_function_call(args)
            merge_imports(self.imports, builtin_handler.required_imports())
            self.last_used_implicits = builtin_handler.get_used_implicits()
        else:
            self.last_used_implicits = self._get_implicits(node.function_name.name)
            args_repr = ", ".join(self.print(x) for x in node.arguments)
            result = f"{fun_repr}({args_repr})"
        if self.last_function:
            self._add_implicits(self.last_function.name, *self.last_used_implicits)
            if self.next_stmt_is_leave:
                return result
            if "termination_token" in self.last_used_implicits:
                n_returns = len(self.last_function.return_variables)
                dummy_returns = ", ".join("Uint256(0, 0)" for _ in range(n_returns))
                return (
                    result
                    + f"\nif termination_token == 1:\nreturn ({dummy_returns})\nend"
                )
        return result

    def visit_expression_statement(self, node: ast.ExpressionStatement) -> str:
        if isinstance(node.expression, ast.FunctionCall):
            return self.visit(node.expression)
        else:
            # see ast.ExpressionStatement docstring
            raise ValueError("What am I going to do with it? Why is it here?..")

    def visit_variable_declaration(self, node: ast.VariableDeclaration) -> str:
        if node.value is None:
            decls_repr = "\n".join(
                self.print(ast.VariableDeclaration(variables=[x], value=ast.Literal(0)))
                for x in node.variables
            )
            return decls_repr
        value_repr = self.visit(node.value)
        vars_repr = ", ".join(f"{self.visit(x)}" for x in node.variables)
        if isinstance(node.value, ast.FunctionCall):
            return f"let ({vars_repr}) = {value_repr}"
        else:
            assert len(node.variables) == 1
            return f"let {vars_repr} = {value_repr}"

    def visit_block(self, node: ast.Block) -> str:
        stmt_reprs = []
        for i, stmt in enumerate(node.statements):
            self.next_stmt_is_leave = False
            if i < len(node.statements) - 1:
                self.next_stmt_is_leave = isinstance(node.statements[i + 1], ast.Leave)
            with self._new_statement():
                stmt_reprs.append(self.visit(stmt))
        return "\n".join(stmt_reprs)

    def visit_function_definition(self, node: ast.FunctionDefinition):
        self.last_function = node
        params_repr = ", ".join(self.print(x) for x in node.parameters)
        returns_repr = ", ".join(self.print(x) for x in node.return_variables)
        body_repr = self.print(node.body)
        implicits = sorted(self.function_to_implicits.setdefault(node.name, set()))
        implicits_decl = ""
        if implicits:
            implicits_decl = "{" + ", ".join(print_implicit(x) for x in implicits) + "}"
        return (
            f"func {node.name}{implicits_decl}({params_repr}) -> ({returns_repr}):\n"
            f"alloc_locals\n"
            f"{body_repr}\n"
            f"end\n"
        )

    def visit_if(self, node: ast.If) -> str:
        cond_repr = self.print(node.condition)
        body_repr = self.print(node.body)
        else_repr = ""
        if node.else_body:
            else_repr = f"else:\n\t{self.print(node.else_body)}\n"
        return (
            f"if {cond_repr}.low + {cond_repr}.high != 0:\n"
            f"\t{body_repr}\n"
            f"{else_repr}"
            f"end"
        )

    def visit_case(self, node: ast.Case):
        return AssertionError("There should be no cases, run SwitchToIfVisitor first")

    def visit_switch(self, node: ast.Switch):
        return AssertionError(
            "There should be no switches, run SwitchToIfVisitor first"
        )

    def visit_for_loop(self, node: ast.ForLoop):
        raise AssertionError(
            "There should be no for loops, run ForLoopEliminator first"
        )

    def visit_break(self, node: ast.Break):
        raise AssertionError("There should be no breaks, run ForLoopEliminator first")

    def visit_continue(self, node: ast.Continue):
        raise AssertionError(
            "There should be no continues, run ForLoopEliminator first"
        )

    def visit_leave(self, _node: ast.Leave) -> str:
        assert self.last_function
        return_names = ", ".join(x.name for x in self.last_function.return_variables)
        return f"return ({return_names})"

    @contextmanager
    def _new_statement(self):
        old = self.last_used_implicits
        self.last_used_implicits = ()
        try:
            yield None
        finally:
            self.last_used_implicits = old

    def _add_implicits(self, fn_name: str, *implicits: str):
        self.function_to_implicits.setdefault(fn_name, set()).update(implicits)

    def _get_implicits(self, fn_name: str) -> tuple[str, ...]:
        return tuple(
            sorted(self.function_to_implicits.setdefault(fn_name, IMPLICITS_SET))
        )

    def _make_constructor(self):
        ctor_implicits = frozenset(self.function_to_implicits.get("__constructor_meat"))
        ctor_implicits |= {"range_check_ptr"}  # for validate_array
        assert ctor_implicits, "The '__constructor_meat' function is not generated"
        manual_implicits = sorted(ctor_implicits & MANUAL_IMPLICITS)
        builtin_implicits = sorted(ctor_implicits - MANUAL_IMPLICITS)

        builtins_str = "{" + ", ".join(map(print_implicit, builtin_implicits)) + "}"
        manuals_str = ", ".join(manual_implicits)
        manuals_init = "".join(map(initialize_manual_implicit, manual_implicits))
        manuals_fin = "".join(map(finalize_manual_implicit, manual_implicits))
        if not manual_implicits:
            meat = "__constructor_meat()"
        else:
            meat = f"with {manuals_str}:\n__constructor_meat()\nend"

        return (
            f"@constructor\n"
            f"func constructor{builtins_str}(calldata_size, calldata_len, calldata : felt*):\n"
            f"alloc_locals\n"
            f"validate_array(calldata_size, calldata_len, calldata)\n"
            f"{manuals_init}"
            f"{meat}\n"
            f"{manuals_fin}"
            f"return ()\n"
            f"end\n"
        )

    def _make_main(self):
        main_implicits = frozenset(self.function_to_implicits.get("__main_meat"))
        main_implicits |= {"range_check_ptr"}  # for validate_array
        assert main_implicits, "The '__main_meat' function is not generated"
        manual_implicits = sorted(main_implicits & MANUAL_IMPLICITS)
        builtin_implicits = sorted(main_implicits - MANUAL_IMPLICITS)

        builtins_str = "{" + ", ".join(map(print_implicit, builtin_implicits)) + "}"
        manuals_str = ", ".join(manual_implicits)
        manuals_init = "".join(map(initialize_manual_implicit, manual_implicits))
        manuals_fin = "".join(map(finalize_manual_implicit, manual_implicits))
        if not manual_implicits:
            meat = "__main_meat()"
        else:
            meat = f"with {manuals_str}:\n__main_meat()\nend"
        if "exec_env" in manual_implicits:
            returns = "(exec_env.to_returndata_size, exec_env.to_returndata_len, exec_env.to_returndata)"
        else:
            returns = "(0, 0, cast(0, felt*))"

        return (
            f"@external\n"
            f"func __main{builtins_str}(calldata_size, calldata_len, calldata : felt*)"
            f"-> (returndata_size, returndata_len, returndata: felt*):\n"
            f"alloc_locals\n"
            f"validate_array(calldata_size, calldata_len, calldata)\n"
            f"{manuals_init}"
            f"{meat}\n"
            f"{manuals_fin}"
            f"return {returns}\n"
            f"end\n"
        )



from contextlib import contextmanager
from functools import reduce
from typing import FrozenSet, List, Sequence, Set, Tuple, Union

import yul.yul_ast as ast
from yul.AstMapper import AstMapper

LiveVars = FrozenSet[str]

fset = frozenset


class DeadcodeEliminator(AstMapper):
    """This class removes any unused variale assignments.

    While this functionality is completely generalisable it
    requires the following passes to be used before it:

    - SwitchToIfVisitor
    - ForLoopIliminator
    """

    def __init__(self):
        self.return_vars: FrozenSet[str] = fset()
        self.live_vars: FrozenSet[str] = fset()

    def map(self, node: ast.Node, **kwargs) -> ast.Node:
        (vs, node) = self.visit(node)
        assert not vs, "There should be no uninstantiated variables in the top scope."
        return node

    def visit_list(
        self, nodes: Sequence[ast.Node], *args, **kwargs
    ) -> Tuple[LiveVars, Tuple[ast.Node]]:
        if len(nodes) == 0:
            return (set(), list())
        lis = [self.visit(x, *args, **kwargs) for x in nodes]
        var_set, nodes = map(list, zip(*lis))
        live_vars = reduce(lambda s1, s2: s1 | s2, var_set, fset())
        return (live_vars, nodes)

    def visit_literal(self, node: ast.Literal) -> Tuple[LiveVars, ast.Literal]:
        return (fset(), node)

    def visit_identifier(self, node: ast.Identifier) -> Tuple[LiveVars, ast.Identifier]:
        return (fset({node.name}), node)

    def visit_typed_name(self, node: ast.TypedName) -> Tuple[LiveVars, ast.TypedName]:
        return (fset({node.name}), node)

    def visit_function_call(
        self, node: ast.FunctionCall
    ) -> Tuple[LiveVars, ast.FunctionCall]:
        (live_vars, arguments) = self.visit(node.arguments)
        return (
            live_vars,
            ast.FunctionCall(function_name=node.function_name, arguments=arguments),
        )

    def visit_expression_statement(
        self, node: ast.ExpressionStatement
    ) -> Tuple[LiveVars, ast.ExpressionStatement]:
        (live_vars, _) = self.visit(node.expression)
        return (live_vars, node)

    def visit_assignment(self, node: ast.Assignment) -> Tuple[LiveVars, ast.Assignment]:
        (live_vars, _) = self.visit(node.value)
        return (live_vars, node)

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration
    ) -> Tuple[LiveVars, ast.VariableDeclaration]:
        if node.value is None:
            return (fset(), node)
        (live_vars, _) = self.visit(node.value)
        return (live_vars, node)

    def visit_function_definition(
        self, node: ast.FunctionDefinition
    ) -> Tuple[LiveVars, ast.FunctionDefinition]:
        return_variables = fset({n.name for n in node.return_variables})
        with self._new_function(return_variables):
            (live_vars, body) = self.visit(node.body)
            live_vars -= fset({arg.name for arg in node.parameters})
            return (
                live_vars,
                ast.FunctionDefinition(
                    name=node.name,
                    parameters=node.parameters,
                    return_variables=node.return_variables,
                    body=body,
                ),
            )

    def visit_if(self, node: ast.If) -> Tuple[LiveVars, ast.If]:
        (live_vars_cond, cond) = self.visit(node.condition)
        (live_vars_body, body) = self.visit_block(node.body)
        (live_vars_else, else_body) = (
            self.visit_block(node.else_body) if node.else_body else (fset(), None)
        )
        return (
            live_vars_cond | live_vars_body | live_vars_else,
            ast.If(condition=cond, body=body, else_body=else_body),
        )

    def visit_switch(self, node: ast.Switch) -> Tuple[LiveVars, ast.Switch]:
        # TODO: Implement these so we can run DeadCodeEliminator at any stage
        assert False, "There should be no switches, run SwitchToIfVisitor first"

    def visit_forloop(self, node: ast.ForLoop) -> Tuple[LiveVars, ast.ForLoop]:
        # TODO: Implement these so we can run DeadCodeEliminator at any stage
        assert False, "There should be no for loops, run ForLoopIliminator first"

    def visit_break(self, node: ast.Break) -> Tuple[LiveVars, ast.Break]:
        # TODO: Implement these so we can run DeadCodeEliminator at any stage
        assert False, "There should be no breaks, run ForLoopIliminator first"
        return (fset(), node)

    def visit_continue(self, node: ast.Continue) -> Tuple[LiveVars, ast.Continue]:
        # TODO: Implement these so we can run DeadCodeEliminator at any stage
        assert False, "There should be no continues, run ForLoopIliminator first"
        return (fset(), node)

    def visit_leave(self, node: ast.Leave) -> Tuple[LiveVars, ast.Leave]:
        return (fset(), node)

    def visit_block(self, node: ast.Block) -> Tuple[LiveVars, ast.Block]:
        if not node.statements:
            return (fset(), node)

        live_vars = self.live_vars

        if any(isinstance(statement, ast.Leave) for statement in node.statements):
            live_vars |= self.return_vars

        statements = []
        for n in reversed(node.statements):
            keep = True
            with self._new_block(live_vars):
                (current_live_vars, statement) = self.visit(n)
                if isinstance(statement, (ast.Assignment, ast.VariableDeclaration)):
                    var_list = _get_variables(statement)
                    if not isinstance(statement.value, ast.FunctionCall):
                        # We under approximate for now, we keep the entire
                        # declaration if any of the declared variable is live.
                        keep = any(var in live_vars for var in var_list)
                    live_vars -= fset(var_list)
                if keep:
                    live_vars |= current_live_vars
                    statements.append(statement)
        statements.reverse()
        return (live_vars, ast.Block(tuple(statements)))

    @contextmanager
    def _new_function(self, return_vars):
        old_returns_vars = self.return_vars
        self.return_vars = return_vars
        try:
            yield None
        finally:
            self.return_vars = old_returns_vars

    @contextmanager
    def _new_block(self, live_vars):
        old_live_vars = self.live_vars
        self.live_vars = live_vars
        try:
            yield None
        finally:
            self.live_vars = old_live_vars


def _get_variables(node: Union[ast.Assignment, ast.VariableDeclaration]):
    if isinstance(node, ast.Assignment):
        return [n.name for n in node.variable_names]
    return [n.name for n in node.variables]



import asyncio
import json
import os
from ast import literal_eval
from enum import Enum
from tempfile import NamedTemporaryFile

import click


class Command(Enum):
    INVOKE = 0
    CALL = 1
    DEPLOY = 2
    STATUS = 3


@click.group()
def warp():
    pass


@warp.command()
@click.argument("file_path", type=click.Path(exists=True))
@click.argument("contract_name")
@click.option("--cairo-output", is_flag=True)
@click.option(
    "--yul-optimizations",
    help="A string input for setting the sequence of optimizer steps",
)
def transpile(file_path, contract_name, cairo_output, yul_optimizations):
    """
    FILE_PATH: Path to your solidity contract\n
    CONTRACT_NAME: Name of the primary contract (non-interface, non-library, non-abstract contract) that you wish to transpile

    Generates a JSON file containing the transpiled contract and relevant information.
    """
    from yul.main import transpile_from_solidity

    path = click.format_filename(file_path)

    if isinstance(yul_optimizations, str):
        output = transpile_from_solidity(file_path, contract_name, yul_optimizations)
    else:
        output = transpile_from_solidity(file_path, contract_name)

    if cairo_output:
        cairo_code = output["cairo_code"]
        with open(f"{file_path[:-4]}.cairo", "w") as cairo_file:
            cairo_file.write(cairo_code)
        path = f"{path[:-4]}.cairo"
    else:
        with open(f"{file_path[:-4]}.json", "w") as f:
            json.dump(output, f)
        path = f"{path[:-4]}.json"
    click.echo(f"The transpilation output has been written to {path}")


@warp.command()
@click.option(
    "--program",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help="the path to the transpiled program JSON file",
)
@click.option("--address", required=True, help="contract address")
@click.option(
    "--function",
    required=True,
    help="the name of the function to invoke, as defined in the Solidity contract",
)
@click.option(
    "--inputs",
    required=True,
    help=(
        "function arguments passed as a python literal enclosed in double quotes. "
        "Either [] or () can be used for grouping as Solidity structs or arrays"
    ),
)
@click.option(
    "--network",
    envvar="STARKNET_NETWORK",
    required=True,
    help="A StarkNet network to use. "
    "Either specify it as the option value or set an environment variable STARKNET_NETWORK.",
)
def invoke(program, address, function, inputs, network):
    """
    Invoke the given function from a contract on StarkNet
    """
    from cli.commands import _invoke

    inputs = literal_eval(inputs)
    contract_base = program[: -len(".json")]
    with open(program, "r") as f:
        program_info = json.load(f)
    asyncio.run(
        _invoke(contract_base, program_info, address, function, inputs, network)
    )


@warp.command()
@click.argument(
    "program", nargs=1, required=True, type=click.Path(exists=True, dir_okay=False)
)
@click.option("--constructor_args", required=False, default="[]")
@click.option(
    "--network",
    envvar="STARKNET_NETWORK",
    required=True,
    help="A StarkNet network to use. "
    "Either specify it as the option value or set an environment variable STARKNET_NETWORK.",
)
def deploy(program, constructor_args, network):
    """
    PROGRAM: path to the transpiled program JSON file.

    Compiles the given Cairo program info file and deploys it to StarkNet.
    Contract's address is printed to stdout.
    """
    from cli.commands import _deploy

    try:
        constructor_args = literal_eval(constructor_args)
    except ValueError:
        pass
    assert program.endswith(".json")
    contract_base = program[: -len(".json")]
    with open(program, "r") as pf:
        program_info = json.load(pf)
    tmp = NamedTemporaryFile(mode="w", delete=False)
    cairo_path = tmp.name
    tmp.write(program_info["cairo_code"])
    tmp.close()
    try:
        asyncio.run(
            _deploy(
                cairo_path,
                contract_base,
                program_info,
                constructor_args,
                network,
            )
        )
    finally:
        os.remove(tmp.name)


@warp.command()
@click.argument("tx_hash", nargs=1, required=True)
def status(tx_hash):
    """
    TX_HASH: The transaction hash printed to stdout after invoking/deployment.\n
    To check the status of the invoke/deployment.
    """
    from cli.commands import _status

    asyncio.run(_status(tx_hash))


def main():
    warp()



IMPLICITS = {
    "memory_dict": "DictAccess*",
    "msize": None,
    "pedersen_ptr": "HashBuiltin*",
    "range_check_ptr": None,
    "syscall_ptr": "felt*",
    "bitwise_ptr": "BitwiseBuiltin*",
    "exec_env": "ExecutionEnvironment*",
    "termination_token": None,
}

IMPLICITS_SET = set(IMPLICITS.keys())

# Implicits that we create manually, they are not built into Cairo
MANUAL_IMPLICITS = {"memory_dict", "msize", "exec_env", "termination_token"}
assert MANUAL_IMPLICITS.issubset(IMPLICITS_SET)


def print_implicit(name):
    type_ = IMPLICITS.get(name, None)
    if type_ is None:
        return name
    else:
        return f"{name}: {type_}"


def copy_implicit(name):
    return f"local {print_implicit(name)} = {name}"


def initialize_manual_implicit(name):
    assert name in MANUAL_IMPLICITS
    if name == "memory_dict":
        return (
            f"let (memory_dict) = default_dict_new(0)\n"
            f"let memory_dict_start = memory_dict\n"
        )
    elif name == "msize":
        return "let msize = 0\n"
    elif name == "exec_env":
        return (
            f"let (__fp__, _) = get_fp_and_pc()\n"
            f"local exec_env_ : ExecutionEnvironment = ExecutionEnvironment("
            f"calldata_size=calldata_size, calldata_len=calldata_len, calldata=calldata,"
            f"returndata_size=0, returndata_len=0, returndata=cast(0, felt*),"
            f"to_returndata_size=0, to_returndata_len=0, to_returndata=cast(0, felt*))\n"
            f"let exec_env : ExecutionEnvironment* = &exec_env_\n"
        )
    elif name == "termination_token":
        return "let termination_token = 0\n"
    assert False, f"Unhandled manual implicit: '{name}'"


def finalize_manual_implicit(name):
    assert name in MANUAL_IMPLICITS
    if name == "memory_dict":
        return "default_dict_finalize(memory_dict_start, memory_dict, 0)\n"
    else:
        return ""



import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor


class AstMapper(AstVisitor):
    def map(self, node: ast.Node, *args, **kwargs) -> ast.Node:
        return self.visit(node, *args, **kwargs)

    def visit_typed_name(self, node: ast.TypedName) -> ast.TypedName:
        return node

    def visit_literal(self, node: ast.Literal) -> ast.Literal:
        return node

    def visit_identifier(self, node: ast.Identifier) -> ast.Identifier:
        return node

    def visit_assignment(self, node: ast.Assignment) -> ast.Assignment:
        return ast.Assignment(
            variable_names=self.visit_list(node.variable_names),
            value=self.visit(node.value),
        )

    def visit_function_call(self, node: ast.FunctionCall) -> ast.FunctionCall:
        return ast.FunctionCall(
            function_name=self.visit(node.function_name),
            arguments=self.visit_list(node.arguments),
        )

    def visit_expression_statement(
        self, node: ast.ExpressionStatement
    ) -> ast.ExpressionStatement:
        return ast.ExpressionStatement(self.visit(node.expression))

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration
    ) -> ast.VariableDeclaration:
        return ast.VariableDeclaration(
            variables=self.visit_list(node.variables),
            value=self.visit(node.value) if node.value is not None else None,
        )

    def visit_block(self, node: ast.Block) -> ast.Block:
        return ast.Block(tuple(self.visit_list(node.statements)))

    def visit_function_definition(
        self, node: ast.FunctionDefinition
    ) -> ast.FunctionDefinition:
        return ast.FunctionDefinition(
            name=node.name,
            parameters=self.visit_list(node.parameters),
            return_variables=self.visit_list(node.return_variables),
            body=self.visit(node.body),
        )

    def visit_if(self, node: ast.If) -> ast.If:
        return ast.If(
            condition=self.visit(node.condition),
            body=self.visit(node.body),
            else_body=self.visit(node.else_body)
            if node.else_body is not None
            else None,
        )

    def visit_case(self, node: ast.Case) -> ast.Case:
        return ast.Case(
            value=self.visit(node.value) if node.value is not None else None,
            body=self.visit(node.body),
        )

    def visit_switch(self, node: ast.Switch) -> ast.Switch:
        return ast.Switch(
            expression=self.visit(node.expression),
            cases=self.visit_list(node.cases),
        )

    def visit_for_loop(self, node: ast.ForLoop) -> ast.ForLoop:
        return ast.ForLoop(
            pre=self.visit(node.pre),
            condition=self.visit(node.condition),
            post=self.visit(node.post),
            body=self.visit(node.body),
        )

    def visit_break(self, node: ast.Break) -> ast.Break:
        return node

    def visit_continue(self, node: ast.Continue) -> ast.Continue:
        return node

    def visit_leave(self, node: ast.Leave) -> ast.Leave:
        return node



from __future__ import annotations

import re

import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor
from yul.WarpException import WarpException


class AstParser:
    def __init__(self, text: str):
        self.lines = text.splitlines()
        if len(self.lines) == 0:
            raise WarpException("Text should not be empty")
        self.pos = 0

    def parse_typed_name(self) -> ast.TypedName:
        tabs = self.get_tabs()
        node_type_name = self.get_word(tabs)
        assert node_type_name == "TypedName:", "This node should be of type TypedNode"
        self.pos += 1

        assert self.get_tabs() == tabs + 1, "Wrong indentation"
        node_name, node_type = self.get_word(tabs + 1).split(":")
        self.pos += 1

        return ast.TypedName(name=node_name, type=node_type)

    def parse_literal(self) -> ast.Literal:
        tabs = self.get_tabs()
        assert self.get_word(tabs).startswith(
            "Literal:"
        ), "This node should be of type Literal"
        value = self.get_word(tabs + 8)
        self.pos += 1

        try:
            value = int(value)
        except ValueError:
            pass

        return ast.Literal(value=value)

    def parse_identifier(self) -> ast.Identifier:
        tabs = self.get_tabs()
        assert self.get_word(tabs).startswith(
            "Identifier:"
        ), "This node should be of type Identifier"
        name = self.get_word(tabs + 11)
        self.pos += 1

        return ast.Identifier(name=name)

    def parse_assignment(self) -> ast.Assignment:
        tabs = self.get_tabs()
        assert (
            self.get_word(tabs) == "Assignment:"
        ), "This node should be of type Assignment"
        self.pos += 1
        assert self.get_word(tabs + 1) == "Variables:"
        self.pos += 1

        variables_list = self.parse_list(tabs + 1, self.parse_identifier)
        assert self.get_word(tabs + 1) == "Value:"
        self.pos += 1

        return ast.Assignment(
            variable_names=variables_list, value=self.parse_expression()
        )

    def parse_function_call(self) -> ast.FunctionCall:
        tabs = self.get_tabs()
        assert (
            self.get_word(tabs) == "FunctionCall:"
        ), "This node should be of type FunctionCall"
        self.pos += 1

        return ast.FunctionCall(
            function_name=self.parse_identifier(),
            arguments=self.parse_list(tabs, self.parse_expression),
        )

    def parse_expression_statement(self) -> ast.Statement:
        tabs = self.get_tabs()
        assert (
            self.get_word(tabs) == "ExpressionStatement:"
        ), "This node should be of type ExpressionStatement"
        self.pos += 1

        return ast.ExpressionStatement(expression=self.parse_expression())

    def parse_variable_declaration(self) -> ast.VariableDeclaration:
        tabs = self.get_tabs()
        assert (
            self.get_word(tabs) == "VariableDeclaration:"
        ), "This node should be of type VariableDeclaration"
        self.pos += 1

        assert self.get_tabs() == tabs + 1
        assert self.get_word(tabs + 1) == "Variables:"
        self.pos += 1
        variables = self.parse_list(tabs + 1, self.parse_typed_name)

        assert self.get_tabs() == tabs + 1
        word = self.get_word(tabs + 1)
        self.pos += 1
        assert word.startswith("Value")
        if word.endswith("None"):
            value = None
        else:
            value = self.parse_expression()

        return ast.VariableDeclaration(variables=variables, value=value)

    def parse_block(self) -> ast.Block:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Block:", "This node should be of type Block"
        self.pos += 1

        return ast.Block(statements=self.parse_list(tabs, self.parse_statement))

    def parse_function_definition(self) -> ast.FunctionDefinition:
        tabs = self.get_tabs()
        assert (
            self.get_word(tabs) == "FunctionDefinition:"
        ), "This node should be of type FunctionDefinition"
        self.pos += 1

        assert self.get_tabs() == tabs + 1 and self.get_word(tabs + 1).startswith(
            "Name:"
        )
        fun_name = self.get_word(tabs + 7)
        self.pos += 1

        assert self.get_tabs() == tabs + 1 and self.get_word(tabs + 1) == "Parameters:"
        self.pos += 1
        params = self.parse_list(tabs + 1, self.parse_typed_name)

        assert (
            self.get_tabs() == tabs + 1
            and self.get_word(tabs + 1) == "Return Variables:"
        )
        self.pos += 1

        returns = self.parse_list(tabs + 1, self.parse_typed_name)

        assert self.get_tabs() == tabs + 1 and self.get_word(tabs + 1) == "Body:"
        self.pos += 1
        body = self.parse_block()

        return ast.FunctionDefinition(
            name=fun_name, parameters=params, return_variables=returns, body=body
        )

    def parse_if(self) -> ast.If:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "If:", "This node should be of type If"
        self.pos += 1

        condition = self.parse_expression()
        body = self.parse_block()
        else_body = None
        if self.get_tabs() > tabs:
            else_body = self.parse_block()

        return ast.If(condition=condition, body=body, else_body=else_body)

    def parse_case(self) -> ast.Case:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Case:", "This node should be of type Case"
        self.pos += 1

        try:
            value = self.parse_literal()
        except AssertionError:
            assert (
                self.get_tabs() == tabs + 1 and self.get_word(tabs + 1) == "Default"
            ), "The value must be a literal or None (when it's the default case)"
            value = None
            self.pos += 1

        return ast.Case(value=value, body=self.parse_block())

    def parse_switch(self) -> ast.Switch:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Switch:", "This node should be of type Switch"
        self.pos += 1

        return ast.Switch(
            expression=self.parse_expression(),
            cases=self.parse_list(tabs, self.parse_case),
        )

    def parse_for_loop(self) -> ast.ForLoop:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "ForLoop:", "This node should be of type ForLoop"
        self.pos += 1

        return ast.ForLoop(
            pre=self.parse_block(),
            condition=self.parse_expression(),
            post=self.parse_block(),
            body=self.parse_block(),
        )

    def parse_break(self) -> ast.Break:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Break", "This node should be of type Break"
        self.pos += 1

        return ast.Break()

    def parse_continue(self) -> ast.Continue:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Continue", "This node should be of type Continue"
        self.pos += 1

        return ast.Continue()

    def parse_leave(self) -> ast.Leave:
        tabs = self.get_tabs()
        assert self.get_word(tabs) == "Leave", "This node should be of type Leave"
        self.pos += 1

        return ast.LEAVE

    def parse_node(self) -> ast.Node:
        tabs = self.get_tabs()
        node_type_name = self.get_word(tabs).split(":")[0]

        parser_name = f"parse_{self.get_name(node_type_name)}"
        parser = getattr(self, parser_name, None)
        if parser is None:
            raise WarpException("Wrong node type name!")
        return parser()

    def parse_statement(self) -> ast.Statement:
        statements = [
            "ExpressionStatement",
            "Assignment",
            "VariableDeclaration",
            "FunctionDefinition",
            "If",
            "Switch",
            "ForLoop",
            "Break",
            "Continue",
            "Leave",
            "Block",
        ]

        tabs = self.get_tabs()
        node_type_name = self.get_word(tabs).split(":")[0]
        assert node_type_name in statements, "Not a valid statement"

        return self.parse_node()

    def parse_expression(self) -> ast.Expression:
        tabs = self.get_tabs()
        node_type_name = self.get_word(tabs).split(":")[0]
        assert node_type_name in [
            "Literal",
            "Identifier",
            "FunctionCall",
        ], "Node type must be an expression"

        return self.parse_node()

    def parse_list(self, tabs, parser):
        items = []
        while self.pos < len(self.lines) and self.get_tabs() > tabs:
            item = parser()
            items.append(item)

        return items

    def get_tabs(self):
        tabs = 0
        if self.pos < len(self.lines):
            for c in self.lines[self.pos]:
                if not c == "\t":
                    break
                tabs += 1
            else:
                raise WarpException(
                    "Lines are not supposed to be filled only with tabs"
                )

        return tabs

    def get_word(self, start: int) -> str:
        return self.lines[self.pos][start:]

    def get_name(self, name):
        name = "_".join(re.findall("[A-Z][^A-Z]*", name))
        return name.lower()


class YulPrinter(AstVisitor):
    def format(self, node: ast.Node, tabs: int = 0) -> str:
        return self.visit(node, tabs)

    def visit_typed_name(self, node: ast.TypedName, tabs: int = 0) -> str:
        return f"{node.name}"

    def visit_literal(self, node: ast.Literal, tabs: int = 0) -> str:
        return f"{node.value}"

    def visit_identifier(self, node: ast.Identifier, tabs: int = 0) -> str:
        return f"{node.name}"

    def visit_assignment(self, node: ast.Assignment, tabs: int = 0) -> str:
        variables = ", ".join(self.visit_list(node.variable_names))
        value = self.visit(node.value, 0)
        return f"{variables} := {value}"

    def visit_function_call(self, node: ast.FunctionCall, tabs: int = 0) -> str:
        name = self.visit(node.function_name)
        args = ", ".join(self.visit_list(node.arguments))
        return f"{name}({args})"

    def visit_expression_statement(
        self, node: ast.ExpressionStatement, tabs: int = 0
    ) -> str:
        return self.visit(node.expression, tabs)

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration, tabs: int = 0
    ) -> str:
        variables = ", ".join(self.visit_list(node.variables))

        value = ""
        if node.value is not None:
            value = f" := {self.visit(node.value)}"

        return f"let {variables}{value}"

    def visit_block(self, node: ast.Block, tabs: int = 0) -> str:
        open_block = "{"
        close_block = "}"
        if self.is_short(node.statements):
            statements = "".join(self.visit_list(node.statements))
            return " ".join([open_block, statements, close_block])

        statements = self.visit_list(node.statements, tabs + 1)
        statements = ["\t" * (tabs + 1) + stmt for stmt in statements]
        statements = "\n".join(statements)
        close_block = "\t" * tabs + close_block
        res = "\n".join([open_block, statements, close_block])

        return res

    def visit_function_definition(
        self, node: ast.FunctionDefinition, tabs: int = 0
    ) -> str:
        parameters = ", ".join(self.visit_list(node.parameters, 0))
        ret_vars = ", ".join(self.visit_list(node.return_variables, 0))
        body = self.visit(node.body, tabs)

        res = f"function {node.name}({parameters})"
        if len(node.return_variables) > 0:
            res += f" -> {ret_vars}"
        res += f" {body}"
        return res

    def visit_if(self, node: ast.If, tabs: int = 0) -> str:
        res = f"if {self.visit(node.condition)} "
        res += self.visit(node.body, tabs)
        if node.else_body is not None:
            res += "\n" + "\t" * tabs + "else "
            res += self.visit(node.else_body, tabs)
        return res

    def visit_case(self, node: ast.Case, tabs: int = 0) -> str:
        res = "\t" * tabs
        if node.value is not None:
            res += f"case {self.visit(node.value)} "
        else:
            res += "default "
        res += self.visit(node.body, tabs)
        return res

    def visit_switch(self, node: ast.Switch, tabs: int = 0) -> str:
        res = f"switch {self.visit(node.expression)}\n"
        res += "\n".join(self.visit_list(node.cases, tabs))
        return res

    def visit_for_loop(self, node: ast.ForLoop, tabs: int = 0) -> str:
        res = "for "
        res += self.visit(node.pre, tabs)
        res += f" {self.visit(node.condition)} "
        res += self.visit(node.post, tabs)
        res += f"\n{self.visit(node.body, tabs)}"
        return res

    def visit_break(self, node: ast.Break, tabs: int = 0) -> str:
        return "break"

    def visit_continue(self, node: ast.Continue, tabs: int = 0) -> str:
        return "continue"

    def visit_leave(self, node: ast.Leave, tabs: int = 0) -> str:
        return "leave"

    def is_short(self, stmts: tuple) -> bool:
        if len(stmts) == 0:
            return True
        return len(stmts) == 1 and type(stmts[0]).__name__ not in [
            "Block",
            "FunctionDefinition",
            "If",
            "Switch",
            "ForLoop",
        ]



from typing import Iterable

import yul.yul_ast as ast
from yul.utils import snakify


def get_children(node: ast.Node) -> Iterable[ast.Node]:
    if isinstance(node, ast.Assignment):
        return node.variable_names + [node.value]
    elif isinstance(node, ast.FunctionCall):
        return node.arguments + [node.function_name]
    elif isinstance(node, ast.ExpressionStatement):
        return [node.expression]
    elif isinstance(node, ast.VariableDeclaration):
        return node.variables + ([] if node.value is None else [node.value])
    elif isinstance(node, ast.Block):
        return node.statements
    elif isinstance(node, ast.FunctionDefinition):
        return node.parameters + node.return_variables + [node.body]
    elif isinstance(node, ast.If):
        return [node.condition, node.body] + (
            [] if node.else_body is None else [node.else_body]
        )
    elif isinstance(node, ast.Case):
        return [node.value, node.body]
    elif isinstance(node, ast.Switch):
        return node.cases + [node.expression]
    elif isinstance(node, ast.ForLoop):
        return [node.pre, node.condition, node.post, node.body]
    else:
        return []


class AstVisitor:
    def __init__(self):
        self.path = []
        # ↑ path of nodes from the AST root down to the current node,
        # inclusively

        def path_decorator(method):
            def new_method(node, *args, **kwargs):
                self.path.append(node)
                res = method(node, *args, **kwargs)
                self.path.pop()
                return res

            return new_method

        for node_type in ast.NODE_TYPES:
            visitor_name = "visit_" + snakify(node_type.__name__)
            method = getattr(self, visitor_name, None)
            if method is None:
                method = self.common_visit
            setattr(self, visitor_name, path_decorator(method))

    def visit(self, node: ast.Node, *args, **kwargs):
        method_name = "visit_" + snakify(type(node).__name__)
        method = getattr(self, method_name, self.common_visit)
        return method(node, *args, **kwargs)

    def common_visit(self, node, *args, **kwargs):
        self.visit_list(get_children(node))

    def visit_list(self, nodes: Iterable[ast.Node], *args, **kwargs) -> list:
        return [self.visit(x, *args, **kwargs) for x in nodes]



from __future__ import annotations

from contextlib import contextmanager
from typing import List, Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.extract_block import extract_block_as_function, extract_rec_block_as_function
from yul.NameGenerator import NameGenerator
from yul.Scope import get_scope
from yul.yul_ast import yul_log_not


class ForLoopEliminator(AstMapper):
    """This class removes for-loops from Yul AST. Loops are replaced by
    recursive functions.

    Furthermore, it replaces breaks and continues with leaves.
    "breaking" from a loop requires an additional condition variable.

    """

    def __init__(self, name_gen: NameGenerator):
        super().__init__()
        self.name_gen = name_gen
        self.body_name: Optional[str] = None
        self.loop_name: Optional[str] = None
        self.break_name: Optional[str] = None
        self.leave_name: Optional[str] = None
        self.has_leave: bool = False
        self.aux_functions: List[ast.FunctionDefinition] = []

    def map(self, node: ast.Node, **kwargs):
        if not isinstance(node, ast.Block):
            return self.visit(node)

        block = self.visit(node)
        return ast.Block(block.statements + tuple(self.aux_functions))

    def visit_for_loop(self, node: ast.ForLoop):
        assert not node.pre.statements, "Loop not simplified"

        with self._new_for_loop():
            assert self.loop_name and self.body_name
            assert self.leave_name and self.break_name

            body = self.visit(node.body)
            post = self.visit(node.post)
            body_fun, body_stmt = extract_block_as_function(
                body, self.body_name, has_leave=self.has_leave
            )

            rec_loop_head = lambda rec: self._make_loop_head(
                node.condition, body_stmt, post.statements, rec
            )
            head_fun, head_stmt = extract_rec_block_as_function(
                rec_loop_head,
                self.loop_name,
                has_leave=True,
                # ↑ we leave if the loop condition is not satisified
            )
            self.aux_functions.extend((body_fun, head_fun))

            leave_id = ast.Identifier(self.leave_name)
            call_statements: ast.Statements
            if leave_id not in get_scope(body).modified_variables:
                call_statements = (head_stmt,)
            else:
                call_statements = (
                    ast.VariableDeclaration(
                        variables=[ast.TypedName(self.leave_name)],
                        value=ast.Literal(False),
                    ),
                    head_stmt,
                    ast.If(condition=leave_id, body=ast.LEAVE_BLOCK),
                )

            return ast.Block(call_statements)

    def visit_break(self, node: ast.Break):
        assert self.break_name
        self.has_leave = True
        return ast.Block(
            (
                ast.Assignment(
                    variable_names=[ast.Identifier(self.break_name)],
                    value=ast.Literal(True),
                ),
                ast.LEAVE,
            )
        )

    def visit_continue(self, node: ast.Continue):
        self.has_leave = True
        return ast.LEAVE

    def visit_leave(self, node: ast.Leave):
        self.has_leave = True
        if not self.leave_name:  # not in a loop
            return node
        return ast.Block(
            (
                ast.Assignment(
                    variable_names=[ast.Identifier(self.leave_name)],
                    value=ast.Literal(True),
                ),
                ast.LEAVE,
            )
        )

    @contextmanager
    def _new_for_loop(self):
        old_names = (self.body_name, self.loop_name, self.break_name, self.leave_name)
        (
            self.loop_name,
            self.body_name,
            self.break_name,
        ) = self.name_gen.make_loop_names()
        self.leave_name = self.name_gen.make_leave_name()
        old_has_leave = self.has_leave
        self.has_leave = False
        try:
            yield None
        finally:
            self.body_name, self.loop_name, self.break_name, self.leave_name = old_names
            self.has_leave = old_has_leave

    def _make_loop_head(
        self,
        condition: ast.Expression,
        body_stmt: ast.Statement,
        post_stmts: list[ast.Statement],
        rec: ast.Statement,
    ) -> ast.Block:
        assert self.break_name and self.leave_name
        break_id = ast.Identifier(self.break_name)
        leave_id = ast.Identifier(self.leave_name)
        modified_vars = get_scope(ast.Block((body_stmt,))).modified_variables
        head_stmts: list[ast.Statement] = []
        if break_id in modified_vars:
            head_stmts.append(
                ast.VariableDeclaration(
                    variables=[ast.TypedName(self.break_name)], value=ast.Literal(False)
                )
            )
        head_stmts.append(ast.If(yul_log_not(condition), ast.LEAVE_BLOCK))
        head_stmts.append(body_stmt)
        if break_id in modified_vars:
            head_stmts.append(ast.If(condition=break_id, body=ast.LEAVE_BLOCK))
        if leave_id in modified_vars:
            head_stmts.append(ast.If(condition=leave_id, body=ast.LEAVE_BLOCK))
        head_stmts += post_stmts
        head_stmts.append(rec)
        return ast.Block(tuple(head_stmts))



from __future__ import annotations

from typing import Union

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.WarpException import WarpException

REDUCERS = {
    "add": lambda x, y: x + y,
    "sub": lambda x, y: x - y,
    "mul": lambda x, y: x * y,
    "div": lambda x, y: x / y if y != 0 else 0,
    "mod": lambda x, y: x % y if y != 0 else 0,
    "exp": lambda x, y: x ** y,
    "not": lambda x: ~x % uint256,
    "eq": lambda x, y: int(x == y),
    "and": lambda x, y: x & y,
    "or": lambda x, y: x | y,
    "xor": lambda x, y: x ^ y,
    "shl": lambda x, y: y << x,
    "shr": lambda x, y: y >> x,
    "addmod": lambda x, y, m: (x + y) % m if m != 0 else 0,
    "mulmod": lambda x, y, m: (x * y) % m if m != 0 else 0,
}


uint256 = 2 ** 256


def overflow_check(res):
    if res > uint256:
        raise WarpException("ConstantFolder detected an overflow")
    return res


class ConstantFolder(AstMapper):
    """This class optimises arithmetic and logical on literals.

    Run VariableInliner before this pass, to maximise the chances of
    optimisation.
    """

    def visit_function_call(
        self, node: ast.FunctionCall
    ) -> Union[ast.Literal, ast.FunctionCall]:
        reduced_args = self.visit_list(node.arguments)
        reducer = REDUCERS.get(node.function_name.name)
        if reducer and all(
            isinstance(arg, ast.Literal) and isinstance(arg.value, int)
            for arg in reduced_args
        ):
            return ast.Literal(
                overflow_check(reducer(*[arg.value for arg in reduced_args]))
            )

        return ast.FunctionCall(
            function_name=self.visit(node.function_name),
            arguments=reduced_args,
        )



from __future__ import annotations

import dataclasses
import json
import os
import re
import shutil
import subprocess
import sys

import yul.yul_ast as ast
from yul.BuiltinHandler import get_default_builtins
from yul.ConstantFolder import ConstantFolder
from yul.DeadcodeEliminator import DeadcodeEliminator
from yul.ExpressionSplitter import ExpressionSplitter
from yul.FoldIf import FoldIf
from yul.ForLoopEliminator import ForLoopEliminator
from yul.ForLoopSimplifier import ForLoopSimplifier
from yul.FunctionGenerator import CairoFunctions, FunctionGenerator
from yul.FunctionPruner import FunctionPruner
from yul.LeaveNormalizer import LeaveNormalizer
from yul.NameGenerator import NameGenerator
from yul.parse_object import parse_to_normalized_ast
from yul.Renamer import MangleNamesVisitor
from yul.RevertNormalizer import RevertNormalizer
from yul.ScopeFlattener import ScopeFlattener
from yul.SwitchToIfVisitor import SwitchToIfVisitor
from yul.ToCairoVisitor import ToCairoVisitor
from yul.utils import get_for_contract
from yul.VariableInliner import VariableInliner
from yul.WarpException import warp_assert

AST_GENERATOR = "kudu"


@dataclasses.dataclass()
class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def transpile_from_solidity(
    sol_src_path, main_contract, optimizers_order="VFoLRFD"
) -> dict:
    sol_src_path_modified = str(sol_src_path)[:-4] + "_marked.sol"
    if not shutil.which(AST_GENERATOR):
        sys.exit(f"Please install {AST_GENERATOR} first")

    try:
        result = subprocess.run(
            [AST_GENERATOR, "--yul-json-ast", sol_src_path, main_contract],
            check=True,
            capture_output=True,
        )
        if result.stderr:
            print(
                bcolors.FAIL
                + bcolors.BOLD
                + result.stderr.decode("utf-8")
                + bcolors.ENDC,
                file=sys.stderr,
            )
    except subprocess.CalledProcessError as e:
        print(
            bcolors.FAIL + bcolors.BOLD + e.stderr.decode("utf-8") + bcolors.ENDC,
            file=sys.stderr,
        )
        raise e

    output = get_for_contract(sol_src_path_modified, main_contract, ["abi"])
    warp_assert(output, f"Couldn't extract {main_contract}'s abi from {sol_src_path}")
    (abi,) = output

    codes = json.loads(result.stdout)
    yul_ast = parse_to_normalized_ast(codes)
    cairo_code = transpile_from_yul(yul_ast, optimizers_order)
    try:
        os.remove(sol_src_path_modified)
    except FileNotFoundError:
        pass  # for reentrancy
    return {"cairo_code": cairo_code, "sol_abi": abi}


def transpile_from_yul(yul_ast: ast.Node, optimizers_order) -> str:
    name_gen = NameGenerator()
    cairo_functions = CairoFunctions(FunctionGenerator())
    builtins = get_default_builtins(cairo_functions)

    pass_order = "FlsFleMSVERSf" + optimizers_order

    ast_passes = {
        "Fls": [ForLoopSimplifier().map],
        "Fle": [ForLoopEliminator(name_gen).map],
        "M": [MangleNamesVisitor().map],
        "S": [SwitchToIfVisitor().map],
        "E": [ExpressionSplitter(name_gen).map],
        "Sf": [ScopeFlattener(name_gen).map],
        "V": [VariableInliner().map, ConstantFolder().map],
        "Fo": [FoldIf().map],
        "L": [LeaveNormalizer().map],
        "R": [RevertNormalizer(builtins).map],
        "F": [FunctionPruner().map],
        "D": [DeadcodeEliminator().map],
    }

    for ast_pass in re.findall(r"[A-Z][a-z]*", pass_order):
        ast_mappers = ast_passes.get(ast_pass) or [lambda ast: ast]
        for ast_mapper in ast_mappers:
            yul_ast = ast_mapper(yul_ast)

    cairo_visitor = ToCairoVisitor(name_gen, cairo_functions, get_default_builtins)

    from starkware.cairo.lang.compiler.parser import parse_file

    return parse_file(cairo_visitor.translate(yul_ast)).format()


def main(argv):
    if len(argv) != 3:
        sys.exit("Supply SOLIDITY-CONTRACT and MAIN-CONTRACT-NAME")
    sol_src_path = argv[1]
    main_contract = argv[2]
    transpile_from_solidity(sol_src_path, main_contract)


if __name__ == "__main__":
    main(sys.argv)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable

from yul.WarpException import warp_assert


@dataclass(frozen=True, order=True)
class StorageVar:
    name: str
    arg_types: tuple[str]
    res_type: str


def generate_getter_body(
    getter_var: str, args: Iterable[str], return_name: str = "res"
) -> str:
    args_repr = ", ".join(args)
    return (
        f"let ({return_name}) = {getter_var}.read({args_repr})\nreturn ({return_name})"
    )


def generate_setter_body(setter_var: str, args: Iterable[str]) -> str:
    args_repr = ", ".join(args)
    return f"{setter_var}.write({args_repr})\nreturn ()\n"


def generate_storage_var_declaration(var: StorageVar) -> str:
    warp_assert(
        all(x == "Uint256" for x in var.arg_types),
        "We don't support storage variables parameterized by types other than Uint256",
    )
    warp_assert(
        var.res_type == "Uint256",
        "We don't support storage variables that return types other than Uint256",
    )
    args_repr = ", ".join(f"arg{i} : {typ}" for i, typ in enumerate(var.arg_types))
    return (
        f"@storage_var\n"
        f"func {var.name}({args_repr}) -> (res: {var.res_type}):\n"
        f"end\n"
    )



from __future__ import annotations

from contextlib import contextmanager


class NameGenerator:
    def __init__(self):
        self.n_loops = -1
        self.n_leaves = -1
        self.n_subexpressions = -1
        self.n_blocks = -1
        self.n_ifs = -1

    def make_loop_names(self) -> tuple[str, str, str]:
        self.n_loops += 1
        return (
            self._prefixed(f"loop_{self.n_loops}"),
            self._prefixed(f"loop_body_{self.n_loops}"),
            self._prefixed(f"break_{self.n_loops}"),
        )

    def make_leave_name(self) -> str:
        self.n_leaves += 1
        return self._prefixed(f"leave_{self.n_leaves}")

    def make_subexpr_name(self) -> str:
        self.n_subexpressions += 1
        return self._prefixed(f"subexpr_{self.n_subexpressions}")

    def make_block_name(self) -> str:
        self.n_blocks += 1
        return self._prefixed(f"block_{self.n_blocks}")

    def make_if_name(self) -> str:
        self.n_ifs += 1
        return self._prefixed(f"if_{self.n_ifs}")

    @contextmanager
    def new_block(self):
        old_n_subexpressions = self.n_subexpressions
        self.n_subexpressions = -1
        try:
            yield None
        finally:
            self.n_subexpressions = old_n_subexpressions

    def _prefixed(self, name: str) -> str:
        return "__warp_" + name



from cli.encoding import get_cairo_calldata, get_ctor_evm_calldata, get_evm_calldata
from starkware.starknet.business_logic.internal_transaction_interface import (
    TransactionExecutionInfo,
)
from starkware.starknet.services.api.contract_definition import ContractDefinition
from starkware.starknet.testing.state import StarknetState


async def invoke_method(
    starknet: StarknetState, program_info: dict, address: str, method: str, *args: list
) -> TransactionExecutionInfo:
    evm_calldata = get_evm_calldata(program_info["sol_abi"], method, args)
    cairo_calldata = get_cairo_calldata(evm_calldata)
    return await starknet.invoke_raw(
        contract_address=address,
        selector="__main",
        calldata=cairo_calldata,
        caller_address=0,
    )


async def deploy_contract(
    starknet: StarknetState,
    program_info: dict,
    contract_definition: ContractDefinition,
    *args: list,
) -> str:
    evm_calldata = get_ctor_evm_calldata(program_info["sol_abi"], args)
    cairo_calldata = get_cairo_calldata(evm_calldata)
    return await starknet.deploy(
        contract_definition=contract_definition, constructor_calldata=cairo_calldata
    )



from __future__ import annotations

from typing import Dict, Optional, Tuple

import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor

CallGraph = Dict[ast.FunctionDefinition, Tuple[ast.FunctionDefinition]]


class CallGraphBuilder(AstVisitor):
    def __init__(self):
        super().__init__()
        # ↓ using second dict instead of set to preserve order
        self.callgraph: dict[ast.FunctionDefinition, dict[str, None]] = {}
        self.last_function: Optional[ast.FunctionDefinition] = None

    def gather(self, block: ast.Block) -> CallGraph:
        self.visit(block)
        return _cleanup_callgraph(self.callgraph)

    def visit_function_call(self, node: ast.FunctionCall):
        self.callgraph[self.last_function][node.function_name.name] = None
        self.common_visit(node)

    def visit_function_definition(self, node: ast.FunctionDefinition):
        self.last_function = node
        self.callgraph[node] = {}
        self.common_visit(node)


def _cleanup_callgraph(
    callgraph: dict[ast.FunctionDefinition, dict[str, None]]
) -> CallGraph:
    defined: dict[str, ast.FunctionDefinition] = {
        fun.name: fun for fun in callgraph.keys()
    }
    return {
        fun: tuple(
            defined[call]
            for call in calls.keys()
            if call in defined and call != fun.name
        )
        for (fun, calls) in callgraph.items()
    }


def build_callgraph(block: ast.Block) -> CallGraph:
    return CallGraphBuilder().gather(block)



from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import yul.yul_ast as ast
from yul.AstMapper import AstMapper

Scope = Dict[str, Optional[ast.Literal]]


class VariableInliner(AstMapper):
    """This class inlines the value of variables by tracking their values
    across declarations and assignments.
    """

    def __init__(self):
        super().__init__()
        self.scope: List[Scope] = []

    def visit_block(self, node: ast.Block) -> ast.Block:
        self.scope.append(dict())
        visited_node = super().visit_block(node)
        self.scope.pop()
        return visited_node

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration
    ) -> ast.VariableDeclaration:
        current_scope = self.scope[-1]
        if isinstance(node.value, ast.Literal):
            assert len(node.variables) == 1
            current_scope[node.variables[0].name] = node.value
        elif isinstance(node.value, ast.Identifier):
            assert len(node.variables) == 1
            current_scope[node.variables[0].name] = self._val_lookup(node.value.name)
        elif node.value is None:
            for var in node.variables:
                current_scope[var.name] = ast.Literal(0)
        else:
            assert isinstance(node.value, ast.FunctionCall)

        return ast.VariableDeclaration(
            variables=node.variables,
            value=self.visit(node.value) if node.value is not None else None,
        )

    def visit_assignment(self, node: ast.Assignment) -> ast.Assignment:
        if isinstance(node.value, ast.Literal):
            assert len(node.variable_names) == 1
            var_name = node.variable_names[0].name
            self._assign_value(var_name, node.value)
        elif isinstance(node.value, ast.Identifier):
            assert len(node.variable_names) == 1
            var_name = node.variable_names[0].name
            ident_value = self._val_lookup(node.value.name)
            self._assign_value(var_name, ident_value)
        else:
            assert isinstance(node.value, ast.FunctionCall)
            for var in node.variable_names:
                # Can't calculate the value, invalidate the old one
                self._assign_value(var.name, None)
        return ast.Assignment(
            variable_names=node.variable_names,
            value=self.visit(node.value),
        )

    def visit_identifier(
        self, node: ast.Identifier
    ) -> Union[ast.Literal, ast.Identifier]:
        return self._val_lookup(node.name) or node

    def _info_lookup(self, var_name: str) -> Optional[Tuple[ast.Literal, Scope]]:
        for scope in reversed(self.scope):
            value = scope.get(var_name)
            if value:
                return (value, scope)
        return None

    def _val_lookup(self, var_name: str) -> Optional[ast.Literal]:
        val_and_scope = self._info_lookup(var_name)
        if val_and_scope:
            return val_and_scope[0]
        return None

    def _scope_lookup(self, var_name: str) -> Optional[Scope]:
        val_and_scope = self._info_lookup(var_name)
        if val_and_scope:
            return val_and_scope[1]
        return None

    def _assign_value(self, var_name: str, value: Optional[ast.Literal]):
        definition_scope = self._scope_lookup(var_name)
        if definition_scope:
            definition_scope[var_name] = None
        self.scope[-1][var_name] = value



from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from typing import Mapping, Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.BuiltinHandler import BuiltinHandler
from yul.top_sort import top_sort_ast

REVERT = ast.FunctionCall(
    function_name=ast.Identifier("revert"), arguments=[ast.Literal(0), ast.Literal(0)]
)
REVERT_STMT = ast.ExpressionStatement(REVERT)


class RevertNormalizer(AstMapper):
    """Prerequisites: LeaveNormalizer"""

    def __init__(self, builtins: Mapping[str, BuiltinHandler]):
        super().__init__()
        self.builtins = builtins
        self.revert_functions: set[str] = set()
        self.terminating_functions: defaultdict[str, bool] = defaultdict(lambda: True)
        # Meaningfully terminating functions are functions that break
        # control flow and return some result. The most basic example
        # is 'return'. In theory, 'revert' should also return some
        # result, but we don't support it. The reason is that it's
        # impossible right now to rollback a transaction while
        # returning something on StarkNet.
        # The default value is 'True', so that when we encounter a
        # function we haven't yet marked as terminating (even if it
        # is), we wouldn't throw it away.
        self.terminates_execution: bool = False
        self.terminates_function: bool = False

    def map(self, node: ast.Node, **kwargs) -> ast.Node:
        if isinstance(node, ast.Block):
            node = top_sort_ast(node)
        return self.visit(node)

    def visit_assignment(self, node: ast.Assignment) -> ast.Statement:
        value = self.visit(node.value)
        if _is_revert(value):
            return REVERT_STMT
        else:
            return ast.Assignment(variable_names=node.variable_names, value=node.value)

    def visit_function_call(self, node: ast.FunctionCall) -> ast.FunctionCall:
        name = node.function_name.name
        if name in self.revert_functions:
            return REVERT
        self.terminates_execution = self._is_function_terminating(name)
        return node

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration
    ) -> ast.Statement:
        value = self.visit(node.value) if node.value else None
        if _is_revert(value):
            return REVERT_STMT
        else:
            return ast.VariableDeclaration(variables=node.variables, value=value)

    def visit_block(self, node: ast.Block) -> ast.Block:
        checkpoint: list[ast.Statement] = []
        new_stmts = []
        for stmt in node.statements:
            with self._new_context():
                new_stmt = self.visit(stmt)
                new_stmts.append(new_stmt)
                if _is_revert(new_stmt):
                    return ast.Block((*checkpoint, REVERT_STMT))
                stmt_terminates_execution = self.terminates_execution
                stmt_terminates_function = self.terminates_function
                if self.terminates_execution or self.terminates_function:
                    checkpoint.extend(new_stmts)
                    new_stmts = []
            self.terminates_execution |= stmt_terminates_execution
            self.terminates_function |= stmt_terminates_function
        checkpoint.extend(new_stmts)
        return ast.Block(tuple(checkpoint))

    def visit_function_definition(
        self, node: ast.FunctionDefinition
    ) -> ast.FunctionDefinition:
        with self._new_context():
            body = self.visit(node.body)
            if _is_revert(body):
                self.revert_functions.add(node.name)
            self.terminating_functions[node.name] = self.terminates_execution
            return ast.FunctionDefinition(
                name=node.name,
                parameters=node.parameters,
                return_variables=node.return_variables,
                body=body,
            )

    def visit_if(self, node: ast.If) -> ast.Statement:
        condition = self.visit(node.condition)
        if _is_revert(condition):
            return ast.ExpressionStatement(REVERT)
        body = self.visit(node.body)
        else_body = self.visit(node.else_body) if node.else_body else None
        if _is_revert(body) and _is_revert(else_body):
            return REVERT_STMT
        return ast.If(condition, body, else_body)

    def visit_leave(self, node: ast.Leave) -> ast.Leave:
        self.terminates_function = True
        return node

    @contextmanager
    def _new_context(self):
        old_terminates_execution = self.terminates_execution
        old_terminates_function = self.terminates_function
        self.terminates_execution = False
        self.terminates_function = False
        try:
            yield None
        finally:
            self.terminates_execution = old_terminates_execution
            self.terminates_function = old_terminates_function

    def _is_function_terminating(self, function_name: str) -> bool:
        handler = self.builtins.get(function_name)
        if handler:
            return handler.is_terminating()
        return self.terminating_functions[function_name]


def _is_revert(node: Optional[ast.Node]) -> bool:
    if isinstance(node, ast.FunctionCall):
        return node.function_name.name == "revert"
    if isinstance(node, ast.ExpressionStatement):
        return _is_revert(node.expression)
    if isinstance(node, ast.Block):
        return len(node.statements) == 1 and _is_revert(node.statements[0])
    return False



from __future__ import annotations

from dataclasses import dataclass
from typing import Union
from weakref import WeakKeyDictionary

import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor


@dataclass
class Scope:
    """'Scope' represents symbols encountered in a particular scope.

    - 'bound_variables' maps names of variables declared in the scope
      to their types.

    - 'read_variables' is a set of undeclared ("free") identifiers
      whose value has been required at some point, but couldn't be
      derived from the scope at that point

    - 'modified_variables' is a set of undeclared ("free") identifiers
      which have been modified at some point, but they weren't
      declared in the scope at that point.

    An intuition for how this groups can be used is the following. If
    this scopes is to be extracted into a separate function,
    'read_variables' is the minimal set of parameters of that
    function. 'modified_variables' is the minimal set of return
    variables of that function.

    NOTE: when there is a 'leave' instruction in the scope, it leads
    to ambiguity regarding 'read_variables'. 'leave' means that a
    function should return (thus, read) values of all of its "return
    variables". If the scope is being computed for an 'ast.Block', the
    set of return variables is unknown.

    """

    bound_variables: dict[str, ast.TypedName]
    read_variables: frozenset[ast.Identifier]
    modified_variables: frozenset[ast.Identifier]


EMPTY_SCOPE = Scope({}, frozenset(), frozenset())


class ScopeResolver(AstVisitor):
    """Gathers information necessary to create a 'Scope'."""

    def __init__(self):
        super().__init__()
        self.bound_variables: dict[str, ast.TypedName] = {}
        self.known_variables: set[ast.Identifier] = set()
        # ↑ variables, whose value is known at this point and doesn't
        # need passing from the outside.
        self.read_variables: list[ast.Identifier] = []
        self.modified_variables: list[ast.Identifier] = []

    def compute_uncached_scope(
        self, node: Union[ast.Block, ast.FunctionDefinition]
    ) -> Scope:
        """Computes 'Scope' for the 'node' without attempting to access
        cache (to avoid infinite recursion).

        """
        if isinstance(node, ast.Block):
            self.common_visit(node)
        else:
            assert isinstance(node, ast.FunctionDefinition)
            # Not visiting return_variables, they are not bound or
            # mentioned semantically.
            self.visit_list(node.parameters)
            self.visit(node.body)

        read = frozenset(self.read_variables)
        modified = frozenset(self.modified_variables)
        return Scope(
            bound_variables=self.bound_variables,
            read_variables=read,
            modified_variables=modified,
        )

    def visit_typed_name(self, node: ast.TypedName):
        self.bound_variables[node.name] = node
        self.known_variables.add(ast.Identifier(node.name))

    def visit_identifier(self, node: ast.Identifier, is_function: bool = False):
        if not is_function:
            self._register_read(node)

    def visit_assignment(self, node: ast.Assignment):
        # assigned vars are registered _after_ the assignment is complete
        self.visit(node.value)
        for var in node.variable_names:
            self._register_modification(var)

    def visit_function_call(self, node: ast.FunctionCall):
        self.visit(node.function_name, is_function=True)
        self.visit_list(node.arguments)

    def visit_variable_declaration(self, node: ast.VariableDeclaration):
        # declared vars are registered _after_ the declaration is complete
        if node.value:
            self.visit(node.value)
        self.visit_list(node.variables)

    def visit_block(self, node: ast.Block):
        for var in get_scope(node).read_variables:
            self._register_read(var)
        for var in get_scope(node).modified_variables:
            self._register_modification(var)

    def visit_function_definition(self, node: ast.FunctionDefinition):
        for var in get_scope(node).read_variables:
            self._register_read(var)
        for var in get_scope(node).modified_variables:
            self._register_modification(var)

    def visit_if(self, node: ast.If):
        self.visit(node.condition)
        scope1 = get_scope(node.body)
        scope2 = get_scope(node.else_body) if node.else_body else EMPTY_SCOPE
        # If a variable has been modified in only one branch, it still
        # means that we need a read access to know it's value after
        # the if. For that reason, we include symmetric difference of
        # modified variables to the set of the read ones.
        new_read = (
            scope1.read_variables
            | scope2.read_variables
            | (scope1.modified_variables ^ scope2.modified_variables)
        )
        new_mod = scope1.modified_variables | scope2.modified_variables
        # only variables that have been modified on both possible code
        # paths are fully determined in the scope
        new_known = scope1.modified_variables & scope2.modified_variables
        for var in new_read:
            self._register_read(var)
        for var in new_mod:
            self._register_modification(var)
        self.known_variables.update(new_known)

    def _register_modification(self, var: ast.Identifier):
        if var not in self.known_variables:
            self.modified_variables.append(var)
            self.known_variables.add(var)

    def _register_read(self, var: ast.Identifier):
        if var not in self.known_variables:
            self.read_variables.append(var)


_scope_cache: WeakKeyDictionary = WeakKeyDictionary()


def get_scope(node: Union[ast.Block, ast.FunctionDefinition]) -> Scope:
    """Retrieves node's 'Scope'.

    Doesn't recompute the scope if it's already been computed. Caching
    is performed using weak references, so it doesn't extend nodes'
    lifetimes.

    """
    scope = _scope_cache.get(node)
    if scope:
        return scope
    scope = ScopeResolver().compute_uncached_scope(node)
    _scope_cache[node] = scope
    return scope



from __future__ import annotations

from typing import Sequence

import yul.yul_ast as ast
from yul.parse import parse_node
from yul.Renamer import Renamer


def parse_to_normalized_ast(codes) -> ast.Node:
    deployment_code = parse_node(codes["deploymentCode"])
    runtime_code = parse_node(codes["runtimeCode"])
    assert isinstance(deployment_code, ast.Block)
    assert isinstance(runtime_code, ast.Block)
    return combine_deployment_and_runtime(deployment_code, runtime_code)


def combine_deployment_and_runtime(
    deployment_code: ast.Block, runtime_code: ast.Block
) -> ast.Block:
    ctor_block, deployment_functions = extract_top_level_code(deployment_code)
    main_block, runtime_functions = extract_top_level_code(runtime_code)
    ctor = ast.FunctionDefinition(
        name="__constructor_meat", parameters=[], return_variables=[], body=ctor_block
    )
    main_ = ast.FunctionDefinition(
        name="__main_meat",
        parameters=[],
        return_variables=[],
        body=main_block,
    )

    deployment_names = {
        x.name for x in deployment_functions if isinstance(x, ast.FunctionDefinition)
    }
    runtime_names = {
        x.name for x in runtime_functions if isinstance(x, ast.FunctionDefinition)
    }
    duplicated = deployment_names & runtime_names
    renamer = Renamer(lambda x: x + "_deployment" if x in duplicated else x)
    renamed_deployment_functions = renamer.visit_list(deployment_functions)
    renamed_ctor = renamer.visit(ctor)
    return ast.Block(
        (renamed_ctor, *renamed_deployment_functions, main_, *runtime_functions)
    )


def extract_top_level_code(
    node: ast.Block,
) -> tuple[ast.Block, Sequence[ast.Statement]]:
    # solidity puts top-level code in the first block
    assert node.statements
    top_level_block, *other_stmts = node.statements
    assert isinstance(top_level_block, ast.Block)
    return top_level_block, other_stmts



import os
import pathlib
import shutil
from io import open
from os import environ, path

import pkg_resources
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(path.join(HERE, "README.md")) as f:
    README = f.read()

# automatically captured required modules for install_requires in requirements.txt and as well as configure dependency links
with open(path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")
install_requires = [
    x.strip()
    for x in all_reqs
    if ("git+" not in x) and (not x.startswith("#")) and (not x.startswith("-"))
]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if "git+" not in x]

setup(
    name="sol-warp",
    description="A Solidity to Cairo Transpiler",
    version="0.5.0",
    package_dir={"": "warp"},
    packages=[
        "cairo-src",
        "cairo-src.evm",
        "cli",
        "yul",
        "bin.linux",
        "bin.macos.10",
        "bin.macos.10.14",
        "bin.macos.11",
    ],  # list of all packages
    include_package_data=True,
    package_data={"": ["*.json", "*.cairo", "kudu"]},
    install_requires=install_requires,
    python_requires=">=3.7",  # any python greater than 3.7
    entry_points="""
        [console_scripts]
        warp=cli.warp_cli:main
    """,
    scripts=["scripts/kudu"],
    author="Nethermind",
    keyword="Ethereum, Layer2, ETH, StarkNet, Nethermind, StarkWare, transpilation, warp, transpiler, cairo",
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="https://github.com/NethermindEth/warp",
    download_url="",
    dependency_links=dependency_links,
    author_email="hello@nethermind.io",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
)



from __future__ import annotations

import json
import os
import re
from typing import List, Optional, Sequence

UPPERCASE_PATTERN = re.compile(r"[A-Z]")

STATEMENT_STRINGS = {
    "ExpressionStatement",
    "Assignment",
    "VariableDeclaration",
    "FunctionDefinition",
    "If",
    "Switch",
    "ForLoop",
    "Break",
    "Continue",
    "Leave",
    "Block",
}

UNSUPPORTED_BUILTINS = [
    "balance",
    "basefee",
    "blockhash",
    "chainid",
    "codecopy",
    "codesize",
    "coinbase",
    "create2",
    "delegatecall",
    "difficulty",
    "extcodecopy",
    "extcodehash",
    "extcodesize",
    "gaslimit",
    "gasprice",
    "invalid",
    "log0",
    "log1",
    "log2",
    "log3",
    "log4",
    "number",
    "origin",
    "pc",
    "selfbalance",
    "selfdestruct",
    "timestamp",
]


def get_low_bits(string: str) -> str:
    return get_low_high(string)[0]


def get_low_high(string: str) -> str:
    try:
        value = int(string)
        high, low = divmod(value, 2 ** 128)
        return f"{low}", f"{high}"
    except ValueError:
        return f"{string}.low", f"{string}.high"


def is_statement(node):
    return type(node).__name__ in STATEMENT_STRINGS


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def snakify(camel_case: str) -> str:
    """ThisCaseWord -> this_case_word"""
    return remove_prefix(
        UPPERCASE_PATTERN.sub(lambda m: f"_{m.group(0).lower()}", camel_case), "_"
    )


def camelize(snake_case: str) -> str:
    """this_case_word -> ThisCaseWord"""
    parts = snake_case.split("_")
    if not parts:
        return snake_case
    if any(x == "" for x in parts):
        raise ValueError(
            f"Can't camelize {snake_case}."
            " It probably contains several consecutive underscores."
        )
    return "".join(x.capitalize() for x in parts)


def clean_path(sol_source):
    if sol_source.startswith("./"):
        return sol_source[2:]
    else:
        return sol_source


def get_kudu_output(args: List[str], sol_source) -> dict:
    args = ",".join(args)
    output_str = os.popen(f"kudu --combined-json {args} {sol_source}").read().strip()
    output = json.loads(output_str)
    for contract in output["contracts"].keys():
        output_str = output_str.replace(contract, os.path.abspath(contract))
    return json.loads(output_str)


def get_public_functions(sol_source: str, main_contract: str) -> list[str]:
    get_source_version(sol_source)
    sol_source = os.path.abspath(clean_path(sol_source))
    hashes = get_kudu_output(["hashes"], sol_source)
    public_functions = set()
    for v in hashes["contracts"][f"{sol_source}:{main_contract}"]["hashes"].keys():
        public_functions.add(f"fun_{v[:v.find('(')]}")
    return list(public_functions)


def get_function_mutabilities(sol_source, main_contract):
    get_source_version(sol_source)
    sol_source = os.path.abspath(clean_path(sol_source))
    function_visibilities = dict()
    abi = get_kudu_output(["abi"], sol_source)
    for v in abi["contracts"][f"{sol_source}:{main_contract}"]["abi"]:
        if v["type"] == "function":
            function_visibilities[v["name"]] = v["stateMutability"]
    return function_visibilities


def get_for_contract(sol_source, target_contract, output_values):
    get_source_version(sol_source)
    compiled = get_kudu_output(output_values, sol_source)
    for contract, values in compiled["contracts"].items():
        name = contract[contract.find(":") + 1 :]
        if name == target_contract:
            return [values[x] for x in output_values]
    return None


def get_source_version(sol_source: str) -> float:
    with open(sol_source) as f:
        src = f.read()
    code_split = src.split("\n")
    for line in code_split:
        if "pragma" in line:
            ver: float = float(line[line.index("0.") + 2 :].replace(";", ""))
            if ver < 7.6:
                raise Exception(
                    "Please use a version of solidity that is at least 0.7.6"
                )
            return ver
    raise Exception("No Solidity version specified in contract")


def cairoize_bytes(bs: bytes, shifted=False) -> tuple(List[int], int):
    """Represent bytes as an array of 128-bit big-endian integers and
    return a number of unused bytes in the last array cell.

    'shifted' indicates if 'bs' should be shifted 12-bytes to the
    right, so as to be suitable as Warp calldata.

    """
    if shifted:
        bs = b"\x00" * 12 + bs
    unused_bytes = -len(bs) % 16
    bs = bs.ljust(len(bs) + unused_bytes, b"\x00")
    arr = [int.from_bytes(bs[i : i + 16], "big") for i in range(0, len(bs), 16)]
    return (arr, unused_bytes)



from __future__ import annotations

from ast import literal_eval

import yul.yul_ast as ast
from yul.utils import camelize, is_statement, remove_prefix
from yul.WarpException import WarpException, warp_assert

_node_to_parser = {}


def register_parser(parser):
    fun_name = parser.__name__
    assert fun_name.startswith("parse_"), (
        "A parsing function should start with 'parse_' "
        "and end with the node type written in snake case"
    )
    node_type = "Yul" + camelize(remove_prefix(fun_name, "parse_"))
    _node_to_parser[node_type] = parser

    def inner(yul_ast):
        warp_assert(
            yul_ast["nodeType"] == node_type,
            f'Expected {node_type}, got {yul_ast["nodeType"]}',
        )

        return parser(yul_ast)

    return inner


def parse_node(yul_ast) -> ast.Node:
    node_type = yul_ast["nodeType"]
    parser = _node_to_parser.get(node_type)
    if parser is None:
        raise WarpException(f"Don't know how to handle {node_type}")
    else:
        return parser(yul_ast)


def parse_expression(yul_ast) -> ast.Expression:
    if yul_ast["nodeType"] == "YulIdentifier":
        return parse_identifier(yul_ast)
    elif yul_ast["nodeType"] == "YulLiteral":
        return parse_literal(yul_ast)
    else:
        return parse_function_call(yul_ast)


def parse_statement(yul_ast) -> ast.Statement:
    node = parse_node(yul_ast)
    warp_assert(
        is_statement(node),
        f"Expected yul_ast.Statement, got {type(node)}",
    )
    return node


@register_parser
def parse_typed_name(yul_ast) -> ast.TypedName:
    name_ = yul_ast["name"]
    type_ = yul_ast["type"]
    if not type_:
        type_ = "Uint256"
    return ast.TypedName(name_, type_)


def read_int(x: str) -> int:
    return int(x, 16) if x.startswith("0x") else int(x)


@register_parser
def parse_literal(yul_ast) -> ast.Literal:
    kind = yul_ast["kind"]
    if kind == "number":
        return ast.Literal(read_int(yul_ast["value"]))
    elif kind == "bool":
        return ast.Literal(yul_ast["value"] == "true")
    elif kind == "string":
        yul_string = yul_ast["hexValue"]
        string_length = len(yul_string)
        encoded_string = read_int("0x" + yul_string + "0" * (64 - string_length))
        return ast.Literal(encoded_string)
    else:
        assert False, "Invalid Literal node"


@register_parser
def parse_identifier(yul_ast) -> ast.Identifier:
    return ast.Identifier(yul_ast["name"])


@register_parser
def parse_assignment(yul_ast) -> ast.Assignment:
    var_names = [parse_identifier(var) for var in yul_ast["variableNames"]]
    value = parse_expression(yul_ast["value"])
    return ast.Assignment(var_names, value)


@register_parser
def parse_function_call(yul_ast) -> ast.FunctionCall:
    fun_name = parse_identifier(yul_ast["functionName"])
    args = [parse_expression(x) for x in yul_ast["arguments"]]
    return ast.FunctionCall(fun_name, args)


@register_parser
def parse_expression_statement(yul_ast) -> ast.ExpressionStatement:
    return ast.ExpressionStatement(parse_expression(yul_ast["expression"]))


@register_parser
def parse_variable_declaration(yul_ast) -> ast.VariableDeclaration:
    variables = [parse_typed_name(x) for x in yul_ast.get("variables", [])]
    value = parse_expression(yul_ast["value"]) if yul_ast["value"] else None
    return ast.VariableDeclaration(variables, value)


@register_parser
def parse_block(yul_ast) -> ast.Block:
    statements = [parse_statement(x) for x in yul_ast["statements"]]
    return ast.Block(statements=statements)


@register_parser
def parse_function_definition(yul_ast) -> ast.FunctionDefinition:
    fun_name = yul_ast["name"]
    params = [parse_typed_name(x) for x in yul_ast.get("parameters", [])]
    returns = [parse_typed_name(x) for x in yul_ast.get("returnVariables", [])]
    body = parse_block(yul_ast["body"])
    return ast.FunctionDefinition(fun_name, params, returns, body)


@register_parser
def parse_if(yul_ast) -> ast.If:
    condition = parse_expression(yul_ast["condition"])
    body = parse_block(yul_ast["body"])
    return ast.If(condition, body)


@register_parser
def parse_case(yul_ast) -> ast.Case:
    return ast.Case(
        parse_literal(yul_ast["value"]) if yul_ast["value"] != "default" else None,
        parse_block(yul_ast["body"]),
    )


@register_parser
def parse_switch(yul_ast) -> ast.Switch:
    return ast.Switch(
        parse_expression(yul_ast["expression"]),
        [parse_case(x) for x in yul_ast["cases"]],
    )


@register_parser
def parse_for_loop(yul_ast) -> ast.ForLoop:
    return ast.ForLoop(
        parse_block(yul_ast["pre"]),
        parse_expression(yul_ast["condition"]),
        parse_block(yul_ast["post"]),
        parse_block(yul_ast["body"]),
    )


@register_parser
def parse_break(yul_ast) -> ast.Break:
    return ast.Break()


@register_parser
def parse_continue(yul_ast) -> ast.Continue:
    return ast.Continue()


@register_parser
def parse_leave(yul_ast) -> ast.Leave:
    return ast.LEAVE



from __future__ import annotations

from typing import Any, Sequence

from eth_abi.codec import ABICodec
from eth_abi.exceptions import EncodingTypeError
from eth_abi.registry import registry as default_registry
from web3._utils.abi import (
    get_abi_input_types,
    get_aligned_abi_inputs,
    get_constructor_abi,
    merge_args_and_kwargs,
)
from web3._utils.contracts import get_function_info
from web3.types import ABI, ABIFunction
from yul.utils import cairoize_bytes


def encode_address(x):
    if not isinstance(x, int):
        raise EncodingTypeError()
    return x.to_bytes(32, byteorder="big")


def decode_address(stream):
    return stream.read(32)


WARP_REGISTRY = default_registry.copy()
WARP_REGISTRY.unregister("address")
WARP_REGISTRY.register("address", encoder=encode_address, decoder=decode_address)
WARP_CODEC = ABICodec(WARP_REGISTRY)


EMPTY_CTOR_ABI = ABIFunction(
    type="constructor",
    inputs=[],
)


def get_ctor_evm_calldata(abi: ABI, inputs: Sequence[Any]) -> bytes:
    ctor_abi = get_constructor_abi(abi) or EMPTY_CTOR_ABI
    arguments = merge_args_and_kwargs(ctor_abi, inputs, {})
    _, aligned_arguments = get_aligned_abi_inputs(ctor_abi, arguments)
    argument_types = get_abi_input_types(ctor_abi)
    encoded_arguments = WARP_CODEC.encode_abi(argument_types, aligned_arguments)
    return encoded_arguments


def get_evm_calldata(abi: ABI, fn_name: str, inputs: Sequence[Any]) -> bytes:
    fn_info = get_function_info(fn_name, WARP_CODEC, contract_abi=abi, args=inputs)
    fn_abi, selector, arguments = fn_info
    argument_types = get_abi_input_types(fn_abi)
    encoded_arguments = WARP_CODEC.encode_abi(argument_types, arguments)
    return bytes.fromhex(selector[2:]) + encoded_arguments


def get_cairo_calldata(evm_calldata: bytes) -> Sequence[int]:
    cairo_input, unused_bytes = cairoize_bytes(evm_calldata, shifted=True)
    calldata_size = (len(cairo_input) * 16) - unused_bytes
    return [calldata_size, len(cairo_input), *cairo_input]
